from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, Mapping, TYPE_CHECKING, cast

from greyhorse.result import Result
from .deps import DepsOperator, DepsProvider
from .operator import OperatorFactoryFn, OperatorFactoryRegistry, OperatorKey
from .providers import ProviderFactoryFn, ProviderKey
from ..errors import NoOpFoundForPattern, OpPolicyViolation
from ..utils.registry import DictRegistry, KeyMapping
from ...i18n import tr
from ...logging import logger

if TYPE_CHECKING:
    from ..schemas.controller import OperatorMappingPolicy

type ControllerKey = type[Controller]
type ControllerFactoryFn = Callable[[...], Result[Controller]]
type ControllerFactoryMapping = Mapping[ControllerKey, ControllerFactoryFn]


class Controller(ABC):
    def __init__(
        self, name: str,
        deps_provider: DepsProvider | None = None,
        deps_operator: DepsOperator | None = None,
    ):
        self._name = name
        self._deps_provider = deps_provider
        self._deps_operator = deps_operator
        self._op_factories = DictRegistry[OperatorKey, OperatorFactoryFn]()

    @property
    def name(self) -> str:
        return self._name

    @property
    @abstractmethod
    def active(self) -> bool:
        ...

    def set_deps_provider(self, instance: DepsProvider):
        self._deps_provider = instance

    def reset_deps_provider(self):
        self._deps_provider = None

    def set_deps_operator(self, instance: DepsOperator):
        self._deps_operator = instance

    def reset_deps_operator(self):
        self._deps_operator = None

    def get_operator_factory(self, key: OperatorKey, name: str | None = None) -> OperatorFactoryFn | None:
        if not self._deps_provider:
            return None
        return self._deps_provider.get_operator_factory(key, name=name)

    def get_provider_factory(self, key: ProviderKey, name: str | None = None) -> ProviderFactoryFn | None:
        if not self._deps_provider:
            return None
        return self._deps_provider.get_provider_factory(key, name=name)

    def get_resource[T](self, key: type, name: str | None = None) -> T | None:
        if not self._deps_provider:
            return None
        if instance := self._deps_provider.get_resource(key, name=name):
            return cast(T, instance)
        return None

    def set_resource(self, key: type, instance: Any, name: str | None = None) -> bool:
        if not self._deps_operator:
            return False
        return self._deps_operator.set_resource(key, instance, name=name)

    def reset_resource(self, key: type, name: str | None = None) -> bool:
        if not self._deps_operator:
            return False
        return self._deps_operator.reset_resource(key, name=name)

    @property
    def operator_factories(self) -> OperatorFactoryRegistry:
        return self._op_factories

    def create(self) -> Awaitable[Result] | Result:
        return Result.from_ok()

    def destroy(self) -> Awaitable[Result] | Result:
        return Result.from_ok()

    def start(self) -> Awaitable[None] | None:
        pass

    def stop(self) -> Awaitable[None] | None:
        pass

    def check_operator_mapping(
        self, policies: list['OperatorMappingPolicy'],
    ) -> Result[Mapping[OperatorKey, KeyMapping[OperatorKey]]]:
        policies_dict = {p.key: p for p in policies}
        key_mapping = {}

        for key in self._op_factories.list_keys():
            if key not in policies_dict:
                error = OpPolicyViolation(type='controller', name=self.name, key=str(key.__name__))
                logger.error(error.message)
                return Result.from_error(error)

            name_pattern = policies_dict[key].name_pattern

            if name_pattern is None:
                key_mapping[key] = KeyMapping[OperatorKey](map_to=policies_dict[key].map_to)
                policies_dict.pop(key)
                continue

            names = []

            for name in self._op_factories.get_names(key):
                if name_pattern.match(name):
                    names.append(name)

            if not names:
                error = NoOpFoundForPattern(
                    type='controller', name=self.name, key=str(key.__name__), pattern=name_pattern,
                )
                logger.error(error.message)
                return Result.from_error(error)

            key_mapping[key] = KeyMapping[OperatorKey](map_to=policies_dict[key].map_to, names=names)
            policies_dict.pop(key)

        for p in policies_dict.values():
            logger.warn(tr('app.entities.operator-unused-policy').format(
                type='controller', name=self.name, key=str(p.key.__name__), pattern=p.name_pattern or '-',
            ))

        return Result.from_ok(key_mapping)



import sys
from itertools import chain

import pydantic

from greyhorse.result import Result
from greyhorse.utils.imports import import_path
from greyhorse.utils.injectors import ParamsInjector
from greyhorse.utils.invoke import invoke_sync
from ..entities.application import Application
from ..entities.controller import Controller, ControllerFactoryFn
from ..entities.module import Module, ModuleErrorsItem
from ..entities.service import Service, ServiceFactoryFn
from ..errors import ControllerCreationError, CtrlFactoryNotFoundError, InvalidModuleConfError, ModuleCreationError, \
    ModuleLoadError, ModuleUnloadError, ModuleValidationError, ServiceCreationError, ServiceFactoryNotFoundError
from ..schemas.controller import ControllerConf
from ..schemas.module import ModuleConf, ModuleDesc
from ..schemas.service import ServiceConf
from ...i18n import tr
from ...logging import logger


def _get_module_package(desc: ModuleDesc) -> str:
    if not desc.path.startswith('.'):
        return desc.path

    path = desc.path
    dots_count = 0

    for c in path[1:]:
        if '.' == c:
            dots_count += 1
        else:
            path = path[dots_count + 1:]
            break

    initpath = desc._initpath
    dots_count = min(dots_count, len(initpath))
    initpath = initpath[0:len(initpath) - dots_count]
    return '.'.join(initpath + [path])


class ModuleBuilder:
    def __init__(
        self, app: Application, root_desc: ModuleDesc,
    ):
        self._app = app
        self._injector = ParamsInjector()
        self._key_stack: list[str] = []
        self._root_desc = root_desc

    def load_pass(self, desc: ModuleDesc | None = None) -> Result[ModuleConf | None]:
        desc = desc or self._root_desc

        if not desc.enabled:
            return Result.from_ok()

        res = self._load_module(desc)
        if not res.success:
            return res

        assert desc._conf
        conf = desc._conf

        if not conf.enabled:
            return Result.from_ok()

        for mod_desc in conf.submodules:
            res = self.load_pass(mod_desc)
            if not res.success:
                return res

        return Result.from_ok(conf)

    def create_module_pass(self, desc: ModuleDesc | None = None) -> Result[Module | None]:
        desc = desc or self._root_desc

        if not desc._conf:
            return Result.from_ok()

        current_conf: ModuleConf = desc._conf

        if not current_conf.enabled:
            logger.warn(tr('app.builder.module-disabled').format(path=desc.path))
            return Result.from_ok()

        logger.info(tr('app.builder.try-create').format(path=desc.path))

        values = {
            'name': current_conf.name,
            'conf': current_conf,
        }

        injected_args = self._injector(current_conf.factory, values=values)

        try:
            instance = invoke_sync(current_conf.factory, *injected_args.args, **injected_args.kwargs)
            assert isinstance(instance, Module)

        except Exception as e:
            error = ModuleCreationError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        if errors := self._create_controllers(current_conf, instance):
            return Result.from_errors(list(chain.from_iterable([e.errors for e in errors])))

        if errors := self._create_services(current_conf, instance):
            return Result.from_errors(list(chain.from_iterable([e.errors for e in errors])))

        self._key_stack.append(current_conf.name)

        for submodule_desc in current_conf.submodules:
            res = self.create_module_pass(submodule_desc)
            if not res.success:
                return res
            if not res.result:
                continue

            submodule = res.result
            # noinspection PyProtectedMember
            submodule._resolve_claims(instance)
            instance.add_module(submodule.name, submodule)

        submodule_key = '.'.join(self._key_stack)
        self._app.register_module(submodule_key, instance, desc)
        self._key_stack.pop()

        logger.info(tr('app.builder.create-success').format(path=desc.path))
        return Result.from_ok(instance)

    def _load_module(self, desc: ModuleDesc) -> Result:
        if desc._conf is not None:
            return Result.from_ok()

        module_path = _get_module_package(desc)
        logger.info(tr('app.builder.try-load').format(module_init_path=module_path))

        try:
            func = import_path(f'{module_path}:__init__')

        except (ImportError, AttributeError) as e:
            error = ModuleLoadError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        injected_args = self._injector(func, values=desc.args)

        try:
            res = invoke_sync(func, *injected_args.args, **injected_args.kwargs)

        except pydantic.ValidationError as e:
            error = ModuleValidationError(detail=str(e))
            logger.error(error.message)
            return Result.from_error(error)
        except Exception as e:
            error = ModuleLoadError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        if not isinstance(res, ModuleConf):
            error = InvalidModuleConfError(module_init_path=module_path)
            logger.error(error.message)
            return Result.from_error(error)

        logger.info(tr('app.builder.load-success').format(module_init_path=module_path))
        desc._conf = res
        return Result.from_ok()

    def _create_controller(
        self, module: Module, conf: ControllerConf, factory: ControllerFactoryFn,
    ) -> Result[Controller]:
        logger.info(tr('app.builder.try-ctrl').format(name=conf.name))
        values = {'name': conf.name, **conf.args}
        injected_args = self._injector(factory, values=values)

        try:
            res = factory(*injected_args.args, **injected_args.kwargs)

        except Exception as e:
            error = ControllerCreationError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        if not res.success:
            return res

        instance = res.result

        res = instance.check_operator_mapping(conf.operator_mapping)
        if not res.success:
            return res

        key_mapping = res.result
        module.add_controller(
            conf.key, instance, name=conf.name,
            key_mapping=key_mapping,
            resources_read=conf.resources_read,
            resources_write=conf.resources_write,
            providers_read=conf.providers_read,
            operators_read=conf.operators_read,
        )

        logger.info(tr('app.builder.ctrl-success').format(name=conf.name))
        return Result.from_ok(instance)

    def _create_controllers(self, module_conf: ModuleConf, instance: Module) -> list[ModuleErrorsItem]:
        errors = []

        for conf in module_conf.controllers:
            if factory := module_conf.controller_factories.get(conf.key):
                res = self._create_controller(instance, conf, factory)
                if not res.success:
                    errors.append(ModuleErrorsItem(where='controller', name=conf.name, errors=res.errors))
            else:
                error = CtrlFactoryNotFoundError(key=str(conf.key.__name__))
                logger.error(error.message)
                errors.append(ModuleErrorsItem(where='controller', name=conf.name, errors=[error]))

        return errors

    def _create_service(
        self, module: Module, conf: ServiceConf, factory: ServiceFactoryFn,
    ) -> Result[Service]:
        logger.info(tr('app.builder.try-service').format(name=conf.name))
        values = {'name': conf.name, **conf.args}
        injected_args = self._injector(factory, values=values)

        try:
            res = factory(*injected_args.args, **injected_args.kwargs)

        except Exception as e:
            error = ServiceCreationError(exc=e)
            logger.error(error.message)
            return Result.from_error(error)

        if not res.success:
            return res

        instance = res.result

        res = instance.check_provider_mapping(conf.provider_mapping)
        if not res.success:
            return res

        prov_key_mapping = res.result

        res = instance.check_operator_mapping(conf.operator_mapping)
        if not res.success:
            return res

        op_key_mapping = res.result

        module.add_service(
            conf.key, instance, name=conf.name,
            op_key_mapping=op_key_mapping,
            prov_key_mapping=prov_key_mapping,
            resources_read=conf.resources_read,
            providers_read=conf.providers_read,
            operators_read=conf.operators_read,
        )

        submodule_key = '.'.join(self._key_stack)
        self._app.register_service(f'{submodule_key}.{conf.name}', instance)

        logger.info(tr('app.builder.service-success').format(name=conf.name))
        return Result.from_ok(instance)

    def _create_services(self, module_conf: ModuleConf, instance: Module) -> list[ModuleErrorsItem]:
        errors = []

        for conf in module_conf.services:
            if factory := module_conf.service_factories.get(conf.key):
                res = self._create_service(instance, conf, factory)
                if not res.success:
                    errors.append(ModuleErrorsItem(where='service', name=conf.name, errors=res.errors))
            else:
                error = ServiceFactoryNotFoundError(key=str(conf.key.__name__))
                logger.error(error.message)
                errors.append(ModuleErrorsItem(where='service', name=conf.name, errors=[error]))

        return errors


class ModuleTerminator:
    def __init__(
        self, app: Application, root_desc: ModuleDesc,
    ):
        self._app = app
        self._injector = ParamsInjector()
        self._key_stack: list[str] = []
        self._root_desc = root_desc

    def unload_pass(self, desc: ModuleDesc | None = None) -> Result:
        desc = desc or self._root_desc

        if not desc.enabled or not desc._conf:
            return Result.from_ok()

        conf = desc._conf

        if not conf.enabled:
            return Result.from_ok()

        for mod_desc in conf.submodules:
            res = self.unload_pass(mod_desc)
            if not res.success:
                return res

        return self._unload_module(desc)

    def destroy_module_pass(
        self, desc: ModuleDesc | None = None, parent: Module | None = None,
    ) -> Result:
        desc = desc or self._root_desc

        if not desc._conf:
            return Result.from_ok()

        logger.info(tr('app.builder.try-destroy').format(path=desc.path))

        current_conf: ModuleConf = desc._conf
        self._key_stack.append(current_conf.name)
        submodule_key = '.'.join(self._key_stack)
        current = self._app.get_module(submodule_key)

        for submodule_desc in current_conf.submodules:
            res = self.destroy_module_pass(submodule_desc, current)
            if not res.success:
                return res

        if parent:
            parent.remove_module(current_conf.name)

        for srv in current.service_list:
            self._app.unregister_service(f'{submodule_key}.{srv.name}')

        self._app.unregister_module(submodule_key)
        self._key_stack.pop()

        logger.info(tr('app.builder.destroy-success').format(path=desc.path))
        return Result.from_ok()

    def _unload_module(self, desc: ModuleDesc) -> Result:
        if desc._conf is None:
            return Result.from_ok()

        module_path = _get_module_package(desc)
        logger.info(tr('app.builder.try-unload').format(module_init_path=module_path))

        try:
            func = import_path(f'{module_path}:__fini__')

        except (ImportError, AttributeError):
            pass
        else:
            injected_args = self._injector(func, types={ModuleConf: desc._conf})

            try:
                invoke_sync(func, *injected_args.args, **injected_args.kwargs)

            except Exception as e:
                error = ModuleUnloadError(exc=e)
                logger.error(error.message)
                return Result.from_error(error)

        logger.info(tr('app.builder.unload-success').format(module_init_path=module_path))
        desc._conf = None
        del sys.modules[module_path]
        return Result.from_ok()



from typing import Any, Callable

from ..utils.registry import ReadonlyRegistry

type Operator = Any
type OperatorKey = type
type OperatorFactoryFn = Callable[[], Operator]
type OperatorFactoryRegistry = ReadonlyRegistry[OperatorKey, OperatorFactoryFn]



from typing import Pattern

from pydantic import BaseModel, Field

from .components import ResourceConf, ResourcePolicy
from ..entities.controller import ControllerKey
from ..entities.operator import OperatorKey


class OperatorMappingPolicy(BaseModel, frozen=True, arbitrary_types_allowed=True):
    key: OperatorKey
    map_to: OperatorKey
    name_pattern: Pattern | None = None


class ControllerConf(ResourceConf[ControllerKey]):
    operator_mapping: list[OperatorMappingPolicy] = Field(default_factory=list)
    resources_write: list[ResourcePolicy] = Field(default_factory=list)



import asyncio
import threading
from collections import defaultdict
from contextlib import AbstractAsyncContextManager, AbstractContextManager, AsyncExitStack, ExitStack
from contextvars import ContextVar
from typing import Any, AsyncContextManager, Awaitable, Callable, ContextManager, Mapping, override
from uuid import uuid4

from greyhorse.utils.invoke import get_asyncio_loop, invoke_async, invoke_sync, is_like_sync_context_manager, \
    is_like_async_context_manager

type FieldFactory[T] = (
    T | Callable[[], Awaitable[T] | T] |
    AbstractContextManager[T] | AbstractAsyncContextManager[T] |
    Callable[[], AbstractContextManager[T]] | Callable[[], AbstractAsyncContextManager[T]]
)


type ContextManagerLike[T] = (
    AbstractContextManager[T] | AbstractAsyncContextManager[T] |
    Callable[[], AbstractContextManager[T]] | Callable[[], AbstractAsyncContextManager[T]]
)


class Context(object):
    __slots__ = (
        '_factory', '_field_factories', '_finalizers',
        '_params', '_object', '_ident', '_parent', '_prev',
    )

    def __init__(
        self, factory: Callable[[...], Any] | None = None,
        fields: Mapping[str, FieldFactory[Any]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
    ):
        self._factory = factory
        self._field_factories = fields or {}
        self._finalizers = finalizers or []
        self._params: dict[str, Any] = {}
        self._object = None
        self._ident = str(uuid4())
        self._parent: Context | None = None
        self._prev: Context | None = None

    @property
    def ident(self) -> str:
        return self._ident

    def __enter__(self) -> Any:
        raise NotImplementedError

    def __exit__(self, exc_type, exc_value, traceback):
        raise NotImplementedError

    async def __aenter__(self) -> Any:
        raise NotImplementedError

    async def __aexit__(self, exc_type, exc_value, traceback):
        raise NotImplementedError


class _ContextStorage(threading.local):
    __slots__ = ('_storage',)

    def __init__(self):
        self._storage: dict[type, list[Context]] = defaultdict(list)

    def add(self, kind: type, instance: Context):
        self._storage[kind].append(instance)

    def remove(self, kind: type, instance: Context):
        self._storage[kind].remove(instance)

    def get_last(self, kind: type) -> Context | None:
        if objects := self._storage.get(kind, []):
            return objects[0]
        return None


_current_context: ContextVar[Context] = ContextVar('_ctx')
_context_storage = _ContextStorage()


def current_context(kind: type | None = None) -> Context | None:
    if kind is None:
        return _current_context.get(None)
    else:
        return _context_storage.get_last(kind)


def current_scope_id(kind: type | None = None) -> str:
    if kind is None:
        if ctx := _current_context.get(None):
            return ctx.ident
    else:
        if ctx := _context_storage.get_last(kind):
            return ctx.ident

    if loop := get_asyncio_loop():
        return f'asyncio:{id(asyncio.current_task(loop))}'
    else:
        return f'thread:{threading.current_thread().ident}'


class SyncContext[T](Context, ContextManager):
    __slots__ = (
        '_sync_stack', '_sub_contexts', '_counter', '_lock',
    )

    def __init__(
        self, factory: Callable[[...], T] | None = None,
        fields: Mapping[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        contexts: list[ContextManagerLike[T]] | None = None,
    ):
        super().__init__(factory, fields, finalizers)
        self._sync_stack = ExitStack()
        self._sub_contexts: list[tuple[ContextManagerLike, str | None]] = []
        self._counter = 0
        self._lock = threading.Lock()

        if contexts:
            for ctx in contexts:
                if is_like_sync_context_manager(ctx):
                    self._sub_contexts.append((ctx, None))

        names_to_remove = []

        for name, factory in self._field_factories.items():
            if is_like_sync_context_manager(factory):
                self._sub_contexts.append((factory, name))
                names_to_remove.append(name)
            elif is_like_async_context_manager(factory):
                names_to_remove.append(name)

        for name in names_to_remove:
            self._field_factories.pop(name)

    def _create(self):
        if self._factory is not None:
            self._object = self._factory(**self._params)

    def _destroy(self):
        del self._object
        self._object = None

    def _nested_enter(self):
        pass

    def _nested_exit(self):
        pass

    @override
    def __enter__(self) -> T:
        with self._lock:
            self._counter += 1
            if 1 != self._counter:
                self._nested_enter()
                return self._object

        self._sync_stack.__enter__()

        for ctx, field in self._sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if value := self._sync_stack.enter_context(ctx):
                if field is not None:
                    self._params[field] = value

        for name, value in self._field_factories.items():
            if callable(value):
                value = invoke_sync(value)
            self._params[name] = value

        self._create()

        self._prev = _current_context.get(None)
        _current_context.set(self)
        self._parent = _context_storage.get_last(type(self._object))
        _context_storage.add(type(self._object), self)
        return self._object

    @override
    def __exit__(self, exc_type, exc_value, traceback):
        with self._lock:
            self._counter = max(self._counter - 1, 0)
            if 0 != self._counter:
                self._nested_exit()
                return

        _context_storage.remove(type(self._object), self)

        try:
            self._destroy()
        except Exception:
            pass

        for name, value in self._field_factories.items():
            self._params.pop(name, None)

        for _, field in self._sub_contexts:
            if field is not None:
                self._params.pop(field, None)

        try:
            self._sync_stack.__exit__(exc_type, exc_value, traceback)
        except Exception:
            pass

        for finalizer in self._finalizers:
            try:
                invoke_sync(finalizer)
            except Exception:
                pass

        _current_context.set(self._prev)
        self._prev = self._parent = None


class AsyncContext[T](Context, AsyncContextManager):
    __slots__ = (
        '_sync_stack', '_async_stack', '_sync_sub_contexts',
        '_async_sub_contexts', '_counter', '_lock',
    )

    def __init__(
        self, factory: Callable[[...], T] | None = None,
        fields: Mapping[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        contexts: list[ContextManagerLike[T]] | None = None,
    ):
        super().__init__(factory, fields, finalizers)
        self._sync_stack = ExitStack()
        self._async_stack = AsyncExitStack()
        self._sync_sub_contexts: list[tuple[ContextManagerLike[T], str | None]] = []
        self._async_sub_contexts: list[tuple[ContextManagerLike[T], str | None]] = []
        self._counter = 0
        self._lock = asyncio.Lock()

        if contexts:
            for ctx in contexts:
                if is_like_async_context_manager(ctx):
                    self._async_sub_contexts.append((ctx, None))
                elif is_like_sync_context_manager(ctx):
                    self._sync_sub_contexts.append((ctx, None))

        names_to_remove = []

        for name, factory in self._field_factories.items():
            if is_like_async_context_manager(factory):
                self._async_sub_contexts.append((factory, name))
                names_to_remove.append(name)
            elif is_like_sync_context_manager(factory):
                self._sync_sub_contexts.append((factory, name))
                names_to_remove.append(name)

        for name in names_to_remove:
            self._field_factories.pop(name)

    async def _create(self):
        if self._factory is not None:
            self._object = self._factory(**self._params)

    async def _destroy(self):
        del self._object
        self._object = None

    async def _nested_enter(self):
        pass

    async def _nested_exit(self):
        pass

    @override
    async def __aenter__(self) -> T:
        async with self._lock:
            self._counter += 1
            if 1 != self._counter:
                await self._nested_enter()
                return self._object

        self._sync_stack.__enter__()
        await self._async_stack.__aenter__()

        for ctx, field in self._sync_sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if value := self._sync_stack.enter_context(ctx):
                if field is not None:
                    self._params[field] = value

        for ctx, field in self._async_sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if value := await self._async_stack.enter_async_context(ctx):
                if field is not None:
                    self._params[field] = value

        for name, value in self._field_factories.items():
            if callable(value):
                value = await invoke_async(value)
            self._params[name] = value

        await self._create()

        self._prev = _current_context.get(None)
        _current_context.set(self)
        self._parent = _context_storage.get_last(type(self._object))
        _context_storage.add(type(self._object), self)
        return self._object

    @override
    async def __aexit__(self, exc_type, exc_value, traceback):
        async with self._lock:
            self._counter = max(self._counter - 1, 0)
            if 0 != self._counter:
                await self._nested_exit()
                return

        _context_storage.remove(type(self._object), self)

        try:
            await self._destroy()
        except Exception:
            pass

        for name, value in self._field_factories.items():
            self._params.pop(name, None)

        for _, field in self._async_sub_contexts:
            if field is not None:
                self._params.pop(field, None)

        for _, field in self._sync_sub_contexts:
            if field is not None:
                self._params.pop(field, None)

        try:
            await self._async_stack.__aexit__(exc_type, exc_value, traceback)
        except Exception:
            pass

        try:
            self._sync_stack.__exit__(exc_type, exc_value, traceback)
        except Exception:
            pass

        for finalizer in self._finalizers:
            try:
                await invoke_async(finalizer)
            except Exception:
                pass

        _current_context.set(self._prev)
        self._prev = self._parent = None


class SyncContextBuilder[T]:
    def __init__(self, factory: Callable[[...], T]):
        self._factory = factory
        self._fields = {}
        self._finalizers = []
        self._contexts = []

    def add_param(self, name: str, value: FieldFactory[T]):
        self._fields[name] = value

    def add_context(self, context: ContextManagerLike[T]):
        self._contexts.append(context)

    def add_finalizer(self, finalizer: Callable[[], Awaitable[None] | None]):
        self._finalizers.append(finalizer)

    def build(self) -> SyncContext[T]:
        return SyncContext[T](
            factory=self._factory, fields=self._fields,
            finalizers=self._finalizers, contexts=self._contexts,
        )


class AsyncContextBuilder[T]:
    def __init__(self, factory: Callable[[...], T]):
        self._factory = factory
        self._fields = {}
        self._finalizers = []
        self._contexts = []

    def add_param(self, name: str, value: FieldFactory[T]):
        self._fields[name] = value

    def add_context(self, context: ContextManagerLike[T]):
        self._contexts.append(context)

    def add_finalizer(self, finalizer: Callable[[], Awaitable[None] | None]):
        self._finalizers.append(finalizer)

    def build(self) -> AsyncContext[T]:
        return AsyncContext[T](
            factory=self._factory, fields=self._fields,
            finalizers=self._finalizers, contexts=self._contexts,
        )



import enum
from abc import ABC, abstractmethod
from contextlib import AbstractAsyncContextManager, AbstractContextManager, asynccontextmanager, contextmanager
from typing import Callable

from greyhorse.result import Result
from ..context import AsyncContext, SyncContext
from ..utils.registry import ReadonlyRegistry

type ProviderKey = type[Provider]
type ProviderFactoryFn = Callable[[], Provider]
type ProviderFactoryRegistry = ReadonlyRegistry[ProviderKey, ProviderFactoryFn]


class BorrowType(str, enum.Enum):
    CONTEXT = 'context'
    SESSION = 'session'
    EXCLUSIVE = 'exclusive'
    SHARED = 'shared'


class Provider:
    type: BorrowType


class SyncContextProvider[T](Provider):
    type = BorrowType.CONTEXT

    def __init__(self, instance: SyncContext[T]):
        self._instance = instance

    def get(self) -> SyncContext[T]:
        return self._instance


class AsyncContextProvider[T](Provider):
    type = BorrowType.CONTEXT

    def __init__(self, instance: AsyncContext[T]):
        self._instance = instance

    def get(self) -> AsyncContext[T]:
        return self._instance


class SyncSessionProvider[T](Provider, ABC):
    type = BorrowType.SESSION

    @contextmanager
    @abstractmethod
    def acquire_session(self, *args, **kwargs) -> AbstractContextManager[T]:
        ...

    def release(self):
        pass


class AsyncSessionProvider[T](Provider, ABC):
    type = BorrowType.SESSION

    @asynccontextmanager
    @abstractmethod
    async def acquire_session(self, *args, **kwargs) -> AbstractAsyncContextManager[T]:
        ...

    async def release(self):
        pass


class SyncExclusiveProvider[T](Provider, ABC):
    type = BorrowType.EXCLUSIVE

    @abstractmethod
    def acquire(self, *args, **kwargs) -> Result[T]:
        ...

    @abstractmethod
    def release(self, instance: T) -> Result:
        ...


class AsyncExclusiveProvider[T](Provider, ABC):
    type = BorrowType.EXCLUSIVE

    @abstractmethod
    async def acquire(self, *args, **kwargs) -> Result[T]:
        ...

    @abstractmethod
    async def release(self, instance: T) -> Result:
        ...


class SyncSharedProvider[ExclR, SharedR](Provider, ABC):
    type = BorrowType.SHARED

    @abstractmethod
    def acquire_exclusive(self, *args, **kwargs) -> Result[ExclR]:
        ...

    @abstractmethod
    def acquire_shared(self, *args, **kwargs) -> Result[SharedR]:
        ...

    @abstractmethod
    def release(self, instance: ExclR | SharedR) -> Result:
        ...


class AsyncSharedProvider[ExclR, SharedR](Provider, ABC):
    type = BorrowType.SHARED

    @abstractmethod
    async def acquire_exclusive(self, *args, **kwargs) -> Result[ExclR]:
        ...

    @abstractmethod
    async def acquire_shared(self, *args, **kwargs) -> Result[SharedR]:
        ...

    @abstractmethod
    async def release(self, instance: ExclR | SharedR) -> Result:
        ...
